# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from dataclasses import field, replace
from typing import TYPE_CHECKING, cast

from redux import (
    BaseAction,
    CompleteReducerResult,
    FinishAction,
    Immutable,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.app import RegisterAppAction, RegisterRegularAppAction
from ubo_app.store.keypad import (
    Key,
    KeypadAction,
    KeypadEvent,
    KeypadKeyPressAction,
    KeypadKeyPressEvent,
)
from ubo_app.store.sound import (
    SoundChangeVolumeAction,
    SoundDevice,
)
from ubo_app.store.status_icons import IconAction

if TYPE_CHECKING:
    from typing_extensions import TypeAlias
    from ubo_gui.menu.types import Menu
    from ubo_gui.page import PageWidget


class MainState(Immutable):
    current_menu: Menu | None = None
    current_application: type[PageWidget] | None = None
    path: list[int] = field(default_factory=list)


class SetMenuPathAction(BaseAction):
    path: list[str]


MainAction: TypeAlias = (
    InitAction
    | FinishAction
    | IconAction
    | KeypadAction
    | RegisterAppAction
    | SetMenuPathAction
)


def main_reducer(  # noqa: C901
    state: MainState | None,
    action: MainAction,
) -> ReducerResult[MainState, SoundChangeVolumeAction, KeypadEvent]:
    if state is None:
        if isinstance(action, InitAction):
            from ._menus import HOME_MENU

            return MainState(current_menu=HOME_MENU)
        raise InitializationActionError

    if isinstance(action, KeypadKeyPressAction):
        actions = []
        if action.key == Key.UP and len(state.path) == 0:
            actions.append(
                SoundChangeVolumeAction(
                    amount=0.05,
                    device=SoundDevice.OUTPUT,
                ),
            )
        if action.key == Key.DOWN and len(state.path) == 0:
            actions.append(
                SoundChangeVolumeAction(
                    amount=-0.05,
                    device=SoundDevice.OUTPUT,
                ),
            )
        return CompleteReducerResult(
            state=state,
            actions=actions,
            events=[
                KeypadKeyPressEvent(key=action.key),
            ],
        )

    if isinstance(action, RegisterAppAction):
        from ubo_gui.menu import Item, is_sub_menu_item, menu_items

        # TODO(sassanh): clone the menu instead of modifying it
        menu = state.current_menu

        main_menu_item: Item = menu_items(menu)[0]
        if not is_sub_menu_item(main_menu_item):
            msg = 'Main menu item is not a `SubMenuItem`'
            raise TypeError(msg)

        container_menu_item: Item
        if isinstance(action, RegisterRegularAppAction):
            container_menu_item = menu_items(main_menu_item['sub_menu'])[0]
        else:
            container_menu_item = menu_items(main_menu_item['sub_menu'])[1]

        if not is_sub_menu_item(container_menu_item):
            msg = 'Settings menu item is not a `SubMenuItem`'
            raise TypeError(msg)

        cast(list[Item], container_menu_item['sub_menu']['items']).append(
            action.menu_item,
        )
        return replace(state, current_menu=menu)

    if isinstance(action, SetMenuPathAction):
        return replace(state, path=action.path)

    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from redux import (
    Action,
    BaseEvent,
    Immutable,
    InitAction,
    InitializationActionError,
    ReducerResult,
)


class IPState(Immutable):
    ...


def reducer(
    state: IPState | None,
    action: Action,
) -> ReducerResult[IPState, Action, BaseEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return IPState()
        raise InitializationActionError
    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

import asyncio
from collections import defaultdict
from socket import AddressFamily
from typing import cast

import psutil
from constants import INTERNET_STATE_ICON_ID, INTERNET_STATE_ICON_PRIORITY
from pythonping import ping
from reducer import IPState
from ubo_gui.menu.types import ActionItem, HeadlessMenu, SubMenuItem

from ubo_app.store import autorun, dispatch
from ubo_app.store.app import RegisterSettingAppAction
from ubo_app.store.status_icons import StatusIconsRegisterAction
from ubo_app.utils.async_ import create_task


@autorun(lambda state: cast(IPState, getattr(state, 'ip', None)))
def get_ip_addresses(_: IPState) -> list[SubMenuItem]:
    ip_addresses_by_interface = defaultdict(list)
    for interface_name, interface_addresses in psutil.net_if_addrs().items():
        for address in interface_addresses:
            if address.family == AddressFamily.AF_INET:
                ip_addresses_by_interface[interface_name].append(address.address)
    return [
        SubMenuItem(
            label=interface_name,
            icon='cable'
            if interface_name.startswith('eth')
            else 'wifi'
            if interface_name.startswith('wlan')
            else 'computer'
            if interface_name.startswith('lo')
            else 'network_node',
            sub_menu=HeadlessMenu(
                title=f'IP Addresses - {interface_name}',
                items=[
                    ActionItem(label=ip_address, icon='lan', action=print)
                    for ip_address in ip_addresses
                ],
            ),
        )
        for interface_name, ip_addresses in ip_addresses_by_interface.items()
    ]


def is_connected() -> bool:
    try:
        response = ping('1.1.1.1', count=1, timeout=1)
        return response.success()
    except OSError:
        return False


async def check_connection() -> bool:
    while True:
        await asyncio.sleep(1)
        if is_connected():
            dispatch(
                StatusIconsRegisterAction(
                    icon='public',
                    priority=INTERNET_STATE_ICON_PRIORITY,
                    id=INTERNET_STATE_ICON_ID,
                ),
            )
        else:
            dispatch(
                StatusIconsRegisterAction(
                    icon='public_off',
                    priority=INTERNET_STATE_ICON_PRIORITY,
                    id=INTERNET_STATE_ICON_ID,
                ),
            )


create_task(check_connection())


IPMainMenu = SubMenuItem(
    label='IP Addresses',
    icon='lan',
    sub_menu=HeadlessMenu(
        title='WiFi Settings',
        items=get_ip_addresses,
    ),
)


def init_service() -> None:
    dispatch(
        [
            RegisterSettingAppAction(
                menu_item=IPMainMenu,
            ),
        ],
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from ubo_app.logging import logger

if TYPE_CHECKING:
    from ubo_gui.menu.types import Item, Menu

SETTINGS_MENU: Menu = {
    'title': 'Settings',
    'items': [],
}

APPS_MENU: Menu = {
    'title': 'Apps',
    'items': [],
}

MAIN_MENU: Menu = {
    'title': 'Main',
    'items': [
        {
            'label': 'Apps',
            'icon': 'apps',
            'sub_menu': APPS_MENU,
        },
        {
            'label': 'Settings',
            'icon': 'settings',
            'sub_menu': SETTINGS_MENU,
        },
        {
            'label': 'About',
            'action': lambda: logger.info('"About" selected!'),
            'icon': 'info',
        },
    ],
}


def notifications_title() -> str:
    from ubo_gui.notification import notification_manager

    return f'Notifications ({notification_manager.unread_count})'


def notifications_menu_items() -> list[Item]:
    from ubo_gui.notification import notification_manager

    return notification_manager.menu_items()


HOME_MENU: Menu = {
    'title': 'Dashboard',
    'items': [
        {
            'label': '',
            'sub_menu': MAIN_MENU,
            'icon': 'menu',
            'is_short': True,
        },
        {
            'label': '',
            'sub_menu': {
                'title': notifications_title,
                'items': notifications_menu_items,
            },
            'color': 'yellow',
            'icon': 'info',
            'is_short': True,
        },
        {
            'label': 'Turn off',
            'action': lambda: logger.info('"Turn off" selected!'),
            'icon': 'power_settings_new',
            'is_short': True,
        },
    ],
}



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from pathlib import Path

from debouncer import DebounceOptions, debounce
from wifi_manager import (
    add_wireless_connection,
    get_connections,
    get_wifi_device,
    get_wifi_device_state,
    request_scan,
)

from ubo_app.logging import logger
from ubo_app.store import dispatch, subscribe_event
from ubo_app.store.app import RegisterSettingAppAction
from ubo_app.store.wifi import (
    ConnectionState,
    WiFiCreateEvent,
    WiFiType,
    WiFiUpdateAction,
    WiFiUpdateRequestAction,
    WiFiUpdateRequestEvent,
)
from ubo_app.utils.async_ import create_task

IS_RPI = Path('/etc/rpi-issue').exists()


def create_wifi_connection(event: WiFiCreateEvent) -> None:
    connection = event.connection
    ssid = connection.ssid
    password = connection.password

    if not password:
        logger.warn('Password is required')
        return

    async def act() -> None:
        await add_wireless_connection(
            ssid=ssid,
            password=password,
            type=connection.type or WiFiType.nopass,
            hidden=connection.hidden,
        )

        logger.info('Result of running `add_wifi`')

        dispatch(
            WiFiUpdateRequestAction(reset=True),
        )

    create_task(act())


@debounce(
    wait=0.5,
    options=DebounceOptions(leading=True, trailing=False, time_window=2),
)
async def update_wifi_list(_: WiFiUpdateRequestEvent | None = None) -> None:
    connections = await get_connections()

    dispatch(
        WiFiUpdateAction(
            connections=connections,
            state=await get_wifi_device_state(),
            current_connection=next(
                (
                    connection
                    for connection in connections
                    if connection.state is ConnectionState.CONNECTED
                ),
                None,
            ),
        ),
    )


def init_service() -> None:
    from pages import main

    async def setup_listeners() -> None:
        wifi_device = await get_wifi_device()
        if not wifi_device:
            return

        async for _ in wifi_device.properties_changed:
            create_task(update_wifi_list())

    create_task(update_wifi_list())
    create_task(setup_listeners())

    dispatch(
        [
            RegisterSettingAppAction(
                menu_item=main.WiFiMainMenu,
            ),
        ],
    )

    subscribe_event(WiFiCreateEvent, create_wifi_connection)
    subscribe_event(WiFiUpdateRequestEvent, lambda _: create_task(request_scan()))
